package lcsb.mapviewer.run;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import lcsb.mapviewer.annotation.services.ModelAnnotator;
import lcsb.mapviewer.common.IProgressUpdater;
import lcsb.mapviewer.converter.ConverterParams;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.persist.ApplicationContextLoader;
import lcsb.mapviewer.persist.DbUtils;
import lcsb.mapviewer.services.interfaces.IExporterService;
import lcsb.mapviewer.services.interfaces.IExporterService.ExporterParameters;
import lcsb.mapviewer.services.utils.data.ExportFileType;

/**
 * This class prepare data for Bell format (further transformation is done by
 * Marek).
 * 
 * @author Piotr Gawron
 * 
 */
public class BellExport {
	/**
	 * Default class logger.
	 */
	private static Logger		 logger	= Logger.getLogger(BellExport.class.getName());

	/**
	 * Object that performs different types of export transformation.
	 */
	@Autowired
	private IExporterService exporter;

	/**
	 * Object used to annotate the model.
	 */
	@Autowired
	private ModelAnnotator	 modelAnnotator;

	/**
	 * Object used to annotate the model.
	 */
	@Autowired
	private DbUtils					 dbUtils;

	/**
	 * Static main method used to run this stand alone code.
	 * 
	 * @param args
	 *          command line arguments
	 */
	public static void main(String[] args) {
		ApplicationContextLoader.loadApplicationContext("consoleApplicationContext.xml");
		BellExport main = new BellExport();
		ApplicationContextLoader.injectDependencies(main);
		main.run();
	}

	/**
	 * Default constructor.
	 */
	public BellExport() {

	}

	/**
	 * Method that export pd map into file that can be exported to bell (by
	 * Marek).
	 * 
	 */
	private void run() {
		CellDesignerXmlParser p = new CellDesignerXmlParser();
		try {
			String filename = PdMapAnnotations.getLastPdFilename();
			String version = PdMapAnnotations.getLastPdVersion();

			// filename = "testFiles/other_full/ASTHMAP18000X8000V1_38.xml";
			// version = "1.1.38";
			dbUtils.createSessionForCurrentThread();
			Model model = p.createModel(new ConverterParams().filename(filename));

			modelAnnotator.performAnnotations(model, new IProgressUpdater() {
				@Override
				public void setProgress(double progress) {
				}
			});

			ExporterParameters params = new IExporterService.ExporterParameters().model(model).//
					fileName("out/bell/" + version + "-reactions.txt").//
					fileType(ExportFileType.TAB_SEPARATED).//
					type(Species.class).//
					moleculeEdges(false);

			exporter.exportReactions(params);

			params = new IExporterService.ExporterParameters().model(model).//
					fileName("out/bell/" + version + "-species.txt").//
					fileType(ExportFileType.TAB_SEPARATED).//
					type(Species.class).//
					// column(ExportColumn.FULL).//
					moleculeEdges(false);
			exporter.exportSpecies(params);

			params = new IExporterService.ExporterParameters().model(model).//
					fileName("out/bell/" + version + "-compartments.txt").//
					fileType(ExportFileType.TAB_SEPARATED).//
					type(Object.class); //
			// column(ExportColumn.FULL);
			exporter.exportCompartments(params);

		} catch (Exception e) {
			logger.error(e.getMessage(), e);
		} finally {
			dbUtils.closeSessionForCurrentThread();
		}
	}

	/**
	 * @return the modelAnnotator
	 * @see #modelAnnotator
	 */
	public ModelAnnotator getModelAnnotator() {
		return modelAnnotator;
	}

	/**
	 * @param modelAnnotator
	 *          the modelAnnotator to set
	 * @see #modelAnnotator
	 */
	public void setModelAnnotator(ModelAnnotator modelAnnotator) {
		this.modelAnnotator = modelAnnotator;
	}
}
