package lcsb.mapviewer.run;

import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;

import lcsb.mapviewer.annotation.services.MiriamConnector;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.exception.InvalidStateException;
import lcsb.mapviewer.converter.InvalidInputDataExecption;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.model.graphics.PolylineData;
import lcsb.mapviewer.model.map.InconsistentModelException;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.modifier.Catalysis;
import lcsb.mapviewer.model.map.reaction.Modifier;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.modelutils.map.ElementUtils;

/**
 * This class is entry point for console tool to convert data from tab separated
 * file a CellDesigner format.
 * 
 * @author Piotr Gawron
 * 
 */
public class ConsoleConverter {

	/**
	 * Default class logger.
	 */
	private final Logger		logger			 = Logger.getLogger(ConsoleConverter.class);

	/**
	 * Options from the input.
	 */
	private RunOptions			inputOptions;

	/**
	 * Connector used for accessing data from miriam registry.
	 */
	private MiriamConnector	mc					 = new MiriamConnector();

	/**
	 * Util class used to access some information about {@link Element}.
	 */
	private ElementUtils		elementUtils = new ElementUtils();

	/**
	 * Default constructor.
	 * 
	 * @param args
	 *          parameters with which the program was run
	 */
	public ConsoleConverter(String[] args) {
		inputOptions = new RunOptions(args);
	}

	/**
	 * Main entry point to the program.
	 * 
	 * @param args
	 *          parameters with which the program was run
	 */
	public static void main(String[] args) {
		// ApplicationContextLoader.loadApplicationContext("consoleApplicationContext.xml");
		ConsoleConverter main = new ConsoleConverter(args);
		// ApplicationContextLoader.injectDependencies(main);
		try {
			main.run();
		} catch (InvalidInputDataExecption e) {
			main.logger.error(e, e);
		}
	}

	/**
	 * This method transform {@link #inputOptions input data}.
	 * 
	 * @throws InvalidInputDataExecption
	 *           thrown when the input data are invalid
	 */
	private void run() throws InvalidInputDataExecption {
		if (inputOptions.isFormat()) {
			inputOptions.printFormat();
		} else if (!inputOptions.isValidInput() || inputOptions.isHelpOption()) {
			inputOptions.printHelp();
		} else {
			logger.debug("Running...");

			String nodesFilename = inputOptions.getNodesFilename();

			Model model = new ModelFullIndexed(null);
			model.setWidth(inputOptions.getModelSize());
			model.setHeight(inputOptions.getModelSize());

			Set<Element> elements = null;
			try {
				elements = getNodes(nodesFilename);
			} catch (IOException e) {
				throw new InvalidInputDataExecption("Problem with input file: " + nodesFilename, e);
			}

			double middleX = inputOptions.getModelSize() / 2;
			double middleY = inputOptions.getModelSize() / 2;

			double size = elements.size();
			double x = 0;

			for (Element alias : model.getElements()) {
				alias.setWidth(inputOptions.getAliasWidth());
				alias.setHeight(inputOptions.getAliasHeight());
				double angle = x / size * 2 * Math.PI;
				alias.setX(middleX + Math.sin(angle) * inputOptions.getRadius());
				alias.setY(middleY + Math.cos(angle) * inputOptions.getRadius());
				model.addElement(alias);
				x++;
			}

			String reactionsFilename = inputOptions.getReactionsFilename();

			List<Reaction> reactions;
			try {
				reactions = getReactions(reactionsFilename, model);
			} catch (IOException e) {
				throw new InvalidInputDataExecption("Problem with input file: " + reactionsFilename, e);
			}
			model.addReactions(reactions);
			String outputFilename = inputOptions.getOutputFilename();

			CellDesignerXmlParser parser = new CellDesignerXmlParser();
			PrintWriter writer = null;
			try {
				writer = new PrintWriter(outputFilename, "UTF-8");
				writer.println(parser.toXml(model));
			} catch (FileNotFoundException e) {
				throw new InvalidInputDataExecption("Problem with output file: " + outputFilename, e);
			} catch (UnsupportedEncodingException e) {
				logger.error(e, e);
			} catch (InconsistentModelException e) {
				throw new InvalidStateException(e);
			} finally {
				writer.close();
			}

		}
	}

	/**
	 * Returns list of {@link Reaction reactions} obtained from input file.
	 * 
	 * @param filename
	 *          file with information about reactions
	 * @param model
	 *          {@link Model} where the data about nodes is stored
	 * @return list of {@link Reaction reactions} obtained from input file
	 * @throws InvalidInputDataExecption
	 *           thrown when the file is invalid
	 * @throws IOException
	 *           thrown when there is a problem with acceesing input file
	 */
	private List<Reaction> getReactions(String filename, Model model) throws InvalidInputDataExecption, IOException {
		List<Reaction> result = new ArrayList<Reaction>();
		Integer identifierColumn = null;
		Integer typeColumn = null;
		Integer reactantsColumn = null;
		Integer modifiersColumn = null;
		Integer productsColumn = null;
		Integer annotationsColumn = null;
		Set<Integer> ids = new HashSet<Integer>();
		BufferedReader br = new BufferedReader(new FileReader(new File(filename)));
		try {
			String line = br.readLine();
			String[] cols = line.split("\t", -1);
			int columns = cols.length;
			for (int i = 0; i < cols.length; i++) {
				if (cols[i].equalsIgnoreCase(RunOptions.IDENTIFIER_REACTIONS_HEADER)) {
					identifierColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.TYPE_REACTIONS_HEADER)) {
					typeColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.REACTANTS_REACTIONS_HEADER)) {
					reactantsColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.MODIFIERS_REACTIONS_HEADER)) {
					modifiersColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.PRODUCTS_REACTIONS_HEADER)) {
					productsColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.ANNOTATIONS_REACTIONS_HEADER)) {
					annotationsColumn = i;
				} else {
					throw new InvalidInputDataExecption("Unknown column name " + cols[i], filename);
				}
			}
			if (identifierColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.IDENTIFIER_REACTIONS_HEADER + "\" defined.", filename);
			} else if (typeColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.TYPE_REACTIONS_HEADER + "\" defined.", filename);
			} else if (reactantsColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.REACTANTS_REACTIONS_HEADER + "\" defined.", filename);
			} else if (modifiersColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.MODIFIERS_REACTIONS_HEADER + "\" defined.", filename);
			} else if (productsColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.PRODUCTS_REACTIONS_HEADER + "\" defined.", filename);
			} else if (annotationsColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.ANNOTATIONS_REACTIONS_HEADER + "\" defined.", filename);
			}
			int row = 2;
			while ((line = br.readLine()) != null) {
				cols = line.split("\t", -1);
				if (cols.length != columns) {
					throw new InvalidInputDataExecption("Invalid number of columns in row " + row, filename);
				}
				String id = cols[identifierColumn];
				if (ids.contains(id)) {
					throw new InvalidInputDataExecption("Id \"" + id + "\" is not unique.", filename);
				}

				String className = cols[typeColumn];
				Class<?> clazz = elementUtils.getClassByName(className);
				if (clazz == null) {
					throw new InvalidInputDataExecption("Unknown element type: " + className, filename);
				}
				List<Product> products = new ArrayList<Product>();
				String productString = cols[productsColumn];
				for (String string : productString.split(",")) {
					if (!string.isEmpty()) {
						Species alias = model.getElementByElementId(string.trim());
						Product product = new Product(alias);
						products.add(product);
					}
				}
				if (products.size() == 0) {
					throw new InvalidInputDataExecption("List of products cannot be empty (row: " + row + ")", filename);
				}

				List<Reactant> reactants = new ArrayList<Reactant>();
				String reactantString = cols[reactantsColumn];
				for (String string : reactantString.split(",")) {
					if (!string.isEmpty()) {
						Species element = model.getElementByElementId(string.trim());
						Reactant product = new Reactant(element);
						reactants.add(product);
					}
				}
				if (reactants.size() == 0) {
					throw new InvalidInputDataExecption("List of reactants cannot be empty (row: " + row + ")", filename);
				}

				List<Modifier> modifiers = new ArrayList<Modifier>();
				String modifierString = cols[modifiersColumn];
				for (String string : modifierString.split(",")) {
					if (!string.isEmpty()) {
						Species element = model.getElementByElementId(string.trim());
						Modifier modifier = new Catalysis(element);
						modifiers.add(modifier);
					}
				}
				Element alias1 = products.get(0).getElement();
				Element alias2 = reactants.get(0).getElement();

				Point2D middle = new Point2D.Double((alias1.getCenterX() + alias2.getCenterX()) / 2, (alias1.getCenterY() + alias2.getCenterY()) / 2);

				Set<MiriamData> annotations = new HashSet<MiriamData>();
				if (annotationsColumn != null) {
					String[] annotationString = cols[annotationsColumn].split(",");
					for (String string : annotationString) {
						if (!string.isEmpty()) {
							try {
								annotations.add(MiriamType.getMiriamByUri(string));
							} catch (InvalidArgumentException e) {
								throw new InvalidInputDataExecption("Invalid miriam urn: " + string, filename);
							}
						}
					}
				}

				try {
					if (!Reaction.class.isAssignableFrom(clazz)) {
						throw new InvalidInputDataExecption("Invalid reaction type type: " + className, filename);
					}
					Reaction reaction = (Reaction) clazz.newInstance();
					for (Product product : products) {
						product.setLine(new PolylineData(middle, product.getElement().getCenter()));
						reaction.addProduct(product);
					}
					for (Reactant reactant : reactants) {
						reactant.setLine(new PolylineData(reactant.getElement().getCenter(), middle));
						reaction.addReactant(reactant);
					}
					for (Modifier modifier : modifiers) {
						modifier.setLine(new PolylineData(middle, modifier.getElement().getCenter()));
						reaction.addModifier(modifier);
					}
					reaction.addMiriamData(annotations);
					reaction.setIdReaction(id);
					result.add(reaction);
				} catch (InstantiationException e) {
					throw new InvalidStateException(e);
				} catch (IllegalAccessException e) {
					throw new InvalidStateException(e);
				}
				row++;
			}
		} finally {
			br.close();
		}
		return result;
	}

	/**
	 * This method parses the data from input file into set of {@link Element
	 * elements}.
	 * 
	 * @param filename
	 *          nam of input file
	 * @return set of {@link Element elements}
	 * @throws InvalidInputDataExecption
	 *           thrown when the input data is invalid
	 * @throws IOException
	 *           thrown when there is a problem with input file
	 */
	private Set<Element> getNodes(String filename) throws InvalidInputDataExecption, IOException {
		ElementUtils elementUtils = new ElementUtils();

		Set<Element> result = new HashSet<>();
		Set<String> ids = new HashSet<>();

		Integer identifierColumn = null;
		Integer typeColumn = null;
		Integer nameColumn = null;
		Integer annotationsColumn = null;

		BufferedReader br = new BufferedReader(new FileReader(new File(filename)));
		try {
			String line = br.readLine();
			String[] cols = line.split("\t", -1);
			int columns = cols.length;
			for (int i = 0; i < cols.length; i++) {
				if (cols[i].equalsIgnoreCase(RunOptions.IDENTIFIER_NODES_HEADER)) {
					identifierColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.TYPE_NODES_HEADER)) {
					typeColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.NAME_NODES_HEADER)) {
					nameColumn = i;
				} else if (cols[i].equalsIgnoreCase(RunOptions.ANNOTATIONS_NODES_HEADER)) {
					annotationsColumn = i;
				} else {
					throw new InvalidInputDataExecption("Unknown column name " + cols[i], filename);
				}
			}
			if (identifierColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.IDENTIFIER_NODES_HEADER + "\" defined.", filename);
			}
			if (typeColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.TYPE_NODES_HEADER + "\" defined.", filename);
			}
			if (nameColumn == null) {
				throw new InvalidInputDataExecption("No \"" + RunOptions.NAME_NODES_HEADER + "\" defined.", filename);
			}
			int row = 2;
			while ((line = br.readLine()) != null) {
				if (line.isEmpty()) {
					continue;
				}
				cols = line.split("\t", -1);
				if (cols.length != columns) {
					throw new InvalidInputDataExecption(
							"Invalid number of columns in row " + row + ". Expected " + columns + ", but " + cols.length + " found", filename);
				}
				String id = cols[identifierColumn];
				if (ids.contains(id)) {
					throw new InvalidInputDataExecption("Id \"" + id + "\" is not unique.", filename);
				}

				String className = cols[typeColumn];
				Class<?> clazz = elementUtils.getClassByName(className);
				if (clazz == null) {
					throw new InvalidInputDataExecption("Unknown element type: " + className, filename);
				}
				String name = cols[nameColumn];
				if (name.isEmpty()) {
					throw new InvalidInputDataExecption("Name cannot be empty (row " + row + ").", filename);
				}
				Set<MiriamData> annotations = new HashSet<MiriamData>();
				if (annotationsColumn != null) {
					String[] annotationString = cols[annotationsColumn].split(",");
					for (String string : annotationString) {
						if (!string.isEmpty()) {
							try {
								annotations.add(MiriamType.getMiriamByUri(string));
							} catch (InvalidArgumentException e) {
								throw new InvalidInputDataExecption("Invalid miriam urn: " + string, filename);
							}
						}
					}
				}

				try {
					if (!Element.class.isAssignableFrom(clazz)) {
						throw new InvalidInputDataExecption("Invliad element type: " + className, filename);
					}
					Element element = (Element) clazz.getConstructor(String.class).newInstance(id);
					element.setName(name);
					element.addMiriamData(annotations);
					result.add(element);
				} catch (ReflectiveOperationException | SecurityException e) {
					throw new InvalidStateException(e);
				}
				row++;
			}
		} finally {
			br.close();
		}

		return result;
	}

	/**
	 * @return the mc
	 * @see #mc
	 */
	public MiriamConnector getMc() {
		return mc;
	}

	/**
	 * @param mc
	 *          the mc to set
	 * @see #mc
	 */
	public void setMc(MiriamConnector mc) {
		this.mc = mc;
	}
}
