package lcsb.mapviewer.run;

import java.util.List;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.log4j.Logger;

import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.modelutils.map.ElementUtils;

/**
 * This class is used to parse and store input data for {@link ConsoleConverter}
 * .
 * 
 * @author Piotr Gawron
 * 
 */
public class RunOptions {
	/**
	 * Default class logger.
	 */
	private final Logger				logger												= Logger.getLogger(RunOptions.class);

	/**
	 * Default size of the model (width/height).
	 */
	private static final int		MODEL_SIZE										= 8000;
	/**
	 * What is the default radius of the circle used for positioning elements.
	 */
	private static final double	RADIUS												= 3500;

	/**
	 * Default width of created alias.
	 */
	private static final int		ALIAS_WIDTH										= 80;

	/**
	 * Default height of created alias.
	 */
	private static final int		ALIAS_HEIGHT									= 20;

	/**
	 * What is the abbreviation of help option.
	 */
	private static final String	HELP_OPTION										= "h";
	/**
	 * What is the abbreviation of nodes file option.
	 */
	private static final String	NODES_OPTION									= "ni";
	/**
	 * What is the abbreviation of reactions file option.
	 */
	private static final String	REACTIONS_OPTION							= "ri";
	/**
	 * What is the abbreviation of output file option.
	 */
	private static final String	OUTPUT_OPTION									= "o";
	/**
	 * What is the abbreviation of printing file format option.
	 */
	private static final String	FORMAT_OPTION									= "f";

	/**
	 * String used for identifier column in nodes file.
	 */
	public static final String	IDENTIFIER_NODES_HEADER				= "IDENTIFIER";
	/**
	 * String used for type column in nodes file.
	 */
	public static final String	TYPE_NODES_HEADER							= "TYPE";
	/**
	 * String used for name column in nodes file.
	 */
	public static final String	NAME_NODES_HEADER							= "NAME";
	/**
	 * String used for annotations column in nodes file.
	 */
	public static final String	ANNOTATIONS_NODES_HEADER			= "ANNOTATIONS";

	/**
	 * Description of nodes file.
	 */
	private static final String	NODES_DESCRIPTION							= "<nodes-file>\n"
																																+ "This file contains information about node elements (chemicals, proteins, etc.) " //
																																+ "that will be included in the output file. The file is tab separeted list of elements " //
																																+ "with the header row. Following columns are obligatory:\n" //
																																+ "* " + IDENTIFIER_NODES_HEADER
																																+ " - unique identifier of the element\n" //
																																+ "* " + TYPE_NODES_HEADER
																																+ " - type of the element. Following types are acceptable: $TYPES\n" //
																																+ "* " + NAME_NODES_HEADER + " - name of the element\n" //
																																+ "* " + ANNOTATIONS_NODES_HEADER + " - comma separated list of annotations in " //
																																+ "Miriam format (http://www.ebi.ac.uk/miriam/main/)\n";	//

	/**
	 * String used for identifier column in reactions file.
	 */
	public static final String	IDENTIFIER_REACTIONS_HEADER		= IDENTIFIER_NODES_HEADER;
	/**
	 * String used for type column in reactions file.
	 */
	public static final String	TYPE_REACTIONS_HEADER					= TYPE_NODES_HEADER;
	/**
	 * String used for reactants column in reactions file.
	 */
	public static final String	REACTANTS_REACTIONS_HEADER		= "REACTANTS";
	/**
	 * String used for modifiers column in reactions file.
	 */
	public static final String	MODIFIERS_REACTIONS_HEADER		= "MODIFIERS";
	/**
	 * String used for products column in reactions file.
	 */
	public static final String	PRODUCTS_REACTIONS_HEADER			= "PRODUCTS";
	/**
	 * String used for annotations column in reactions file.
	 */
	public static final String	ANNOTATIONS_REACTIONS_HEADER	= ANNOTATIONS_NODES_HEADER;

	/**
	 * Description of reactions file.
	 */
	private static final String	REACTIONS_DESCRIPTION					= "<reactions-file>\n" //
																																+ "This file contains information about reactions. The file is tab separeted list of all " //
																																+ "reactions with the header row. Following columns are obligatory:\n" //
																																+ "* " + IDENTIFIER_REACTIONS_HEADER
																																+ " - unique identifier of the reaction\n" //
																																+ "* "
																																+ TYPE_REACTIONS_HEADER
																																+ " - type of the element. Following types are acceptable: $R_TYPES\n" //
																																+ "* " + REACTANTS_REACTIONS_HEADER + " - comma separated list of reactant identifiers\n" //
																																+ "* " + MODIFIERS_REACTIONS_HEADER + " - comma separated list of modifier identifiers\n" //
																																+ "* " + PRODUCTS_REACTIONS_HEADER + " - comma separated list of product identifiers\n" //
																																+ "* " + ANNOTATIONS_REACTIONS_HEADER + " - comma separated list of annotations in " //
																																+ "Miriam format (http://www.ebi.ac.uk/miriam/main/)\n";	//

	/**
	 * Object with available input options.
	 */
	private Options							options;

	/**
	 * What should be the command line to run the program.
	 */
	private String							commandLineRun								= null;

	/**
	 * Was the input valid?
	 */
	private boolean							validInput										= false;

	/**
	 * Data parsed from input.
	 */
	private CommandLine					commandLine;

	/**
	 * Default constructor that parses data.
	 * 
	 * @param args
	 *          parameters used to run the program
	 */
	public RunOptions(String[] args) {
		options = new Options();
		options.addOption(createOption("nodes-input", NODES_OPTION, true, "nodes-file", "tab separated file with input nodes (chemicals, proteins, etc.)"));
		options.addOption(createOption("reaction-input", REACTIONS_OPTION, true, "reactions-file", "tab separated file with input reactions"));
		options.addOption(createOption("output", OUTPUT_OPTION, true, "output-file", "CellDesigner file with output diagram"));
		OptionBuilder.isRequired(false);
		options.addOption(HELP_OPTION, false, "print this help menu");
		options.addOption(FORMAT_OPTION, false, "print format of input files");

		commandLineRun = "java -jar " + new java.io.File(ConsoleConverter.class.getProtectionDomain().getCodeSource().getLocation().getPath()).getName()
				+ " [options]";

		CommandLineParser parser = new BasicParser();

		try {
			validInput = true;
			commandLine = parser.parse(options, args);
			if (getNodesFilename() == null || getReactionsFilename() == null || getOutputFilename() == null) {
				validInput = false;
			}
		} catch (ParseException e) {
			validInput = false;
			logger.info(e.getMessage());
		}

	}

	/**
	 * Creates new {@link Option} object.
	 * 
	 * @param optionName
	 *          full name of the option
	 * @param optionAbbreviation
	 *          abbreviation of the option
	 * @param hasArgument
	 *          has the option argument
	 * @param argumentName
	 *          name of the argument
	 * @param description
	 *          description of the option
	 * @return {@link Option} created from input values
	 */
	private Option createOption(String optionName, String optionAbbreviation, boolean hasArgument, String argumentName, String description) {
		OptionBuilder.hasArg(hasArgument);
		if (hasArgument) {
			OptionBuilder.withArgName(argumentName);
		}
		OptionBuilder.withDescription(description);
		OptionBuilder.withLongOpt(optionName);
		return OptionBuilder.create(optionAbbreviation);
	}

	/**
	 * @return the options
	 * @see #options
	 */
	public Options getOptions() {
		return options;
	}

	/**
	 * @param options
	 *          the options to set
	 * @see #options
	 */
	public void setOptions(Options options) {
		this.options = options;
	}

	/**
	 * @return the commandLineRun
	 * @see #commandLineRun
	 */
	public String getCommandLineRun() {
		return commandLineRun;
	}

	/**
	 * @param commandLineRun
	 *          the commandLineRun to set
	 * @see #commandLineRun
	 */
	public void setCommandLineRun(String commandLineRun) {
		this.commandLineRun = commandLineRun;
	}

	/**
	 * @return the validInput
	 * @see #validInput
	 */
	public boolean isValidInput() {
		return validInput;
	}

	/**
	 * @param validInput
	 *          the validInput to set
	 * @see #validInput
	 */
	public void setValidInput(boolean validInput) {
		this.validInput = validInput;
	}

	/**
	 * @return the commandLine
	 * @see #commandLine
	 */
	public CommandLine getCommandLine() {
		return commandLine;
	}

	/**
	 * @param commandLine
	 *          the commandLine to set
	 * @see #commandLine
	 */
	public void setCommandLine(CommandLine commandLine) {
		this.commandLine = commandLine;
	}

	/**
	 * Prints help to the console how to used the program (what parameters to put
	 * and what are they used for).
	 */
	public void printHelp() {
		HelpFormatter formatter = new HelpFormatter();
		formatter.printHelp(commandLineRun, options);
	}

	/**
	 * Prints information about input file formats.
	 */
	public void printFormat() {
		String desc = NODES_DESCRIPTION;
		ElementUtils eu = new ElementUtils();
		List<Class<? extends Element>> classes = eu.getAvailableElementSubclasses();
		String classString = "";
		for (Class<? extends Element> class1 : classes) {
			classString += class1.getSimpleName() + ", ";
		}
		desc = desc.replaceAll("\\$TYPES", classString);
		desc += "\n";

		List<Class<? extends Reaction>> rClasses = eu.getAvailableReactionSubclasses();
		classString = "";
		for (Class<? extends Reaction> class1 : rClasses) {
			classString += class1.getSimpleName().replace("Reaction", "") + ", ";
		}

		desc += REACTIONS_DESCRIPTION.replaceAll("\\$R_TYPES", classString);
		System.out.println(desc);
	}

	/**
	 * Checks if {@link #HELP_OPTION} was selected in the input.
	 * 
	 * @return <code>true</code> if {@link #HELP_OPTION} was selected,
	 *         <code>false</code> otherwise
	 */
	public boolean isHelpOption() {
		return commandLine.hasOption(HELP_OPTION);
	}

	/**
	 * Returns name of the nodes file.
	 * 
	 * @return name of the nodes file
	 */
	public String getNodesFilename() {
		return commandLine.getOptionValue(NODES_OPTION);
	}

	/**
	 * Returns name of the reactions file.
	 * 
	 * @return name of the reactions file
	 */
	public String getReactionsFilename() {
		return commandLine.getOptionValue(REACTIONS_OPTION);
	}

	/**
	 * Returns name of the output file.
	 * 
	 * @return name of the output file
	 */
	public String getOutputFilename() {
		return commandLine.getOptionValue(OUTPUT_OPTION);
	}

	/**
	 * Checks if {@link #FORMAT_OPTION} was selected in the input.
	 * 
	 * @return <code>true</code> if {@link #FORMAT_OPTION} was selected,
	 *         <code>false</code> otherwise
	 */
	public boolean isFormat() {
		return commandLine.hasOption(FORMAT_OPTION);
	}

	/**
	 * Returns size of created {@link lcsb.mapviewer.model.map.model.Model}.
	 * 
	 * @return size of created {@link lcsb.mapviewer.model.map.model.Model}
	 */
	public int getModelSize() {
		return RunOptions.MODEL_SIZE;
	}

	/**
	 * Returns width of created
	 * {@link lcsb.mapviewer.model.map.species.Element}.
	 * 
	 * @return width of created
	 *         {@link lcsb.mapviewer.model.map.species.Element}
	 */
	public int getAliasWidth() {
		return RunOptions.ALIAS_WIDTH;
	}

	/**
	 * Returns height of created
	 * {@link lcsb.mapviewer.model.map.species.Element}.
	 * 
	 * @return height of created
	 *         {@link lcsb.mapviewer.model.map.species.Element}
	 */
	public int getAliasHeight() {
		return RunOptions.ALIAS_HEIGHT;
	}

	/**
	 * Returns radius of the circle on which elements will be located in the
	 * output CellDesigner file.
	 * 
	 * @return radius of the circle on which elements will be located
	 */
	public double getRadius() {
		return RunOptions.RADIUS;
	}

}
